/* Do not modify this file.  */
/* It is created automatically by the Makefile.  */
#include "register.h"
void register_all_tap_listeners(void) {
  { extern void register_tap_listener_gtkafpstat (void);
    register_tap_listener_gtkafpstat ();}
  { extern void register_tap_listener_gtkansi_a_stat (void);
    register_tap_listener_gtkansi_a_stat ();}
  { extern void register_tap_listener_gtkansi_map_stat (void);
    register_tap_listener_gtkansi_map_stat ();}
  { extern void register_tap_listener_gtkdhcpstat (void);
    register_tap_listener_gtkdhcpstat ();}
  { extern void register_tap_listener_gtk_camelcounter (void);
    register_tap_listener_gtk_camelcounter ();}
  { extern void register_tap_listener_gtk_camelsrt (void);
    register_tap_listener_gtk_camelsrt ();}
  { extern void register_tap_listener_gtkcomparestat (void);
    register_tap_listener_gtkcomparestat ();}
  { extern void register_tap_listener_eth_conversation (void);
    register_tap_listener_eth_conversation ();}
  { extern void register_tap_listener_fc_conversation (void);
    register_tap_listener_fc_conversation ();}
  { extern void register_tap_listener_fddi_conversation (void);
    register_tap_listener_fddi_conversation ();}
  { extern void register_tap_listener_ip_conversation (void);
    register_tap_listener_ip_conversation ();}
  { extern void register_tap_listener_ipx_conversation (void);
    register_tap_listener_ipx_conversation ();}
  { extern void register_tap_listener_jxta_conversation (void);
    register_tap_listener_jxta_conversation ();}
  { extern void register_tap_listener_ncp_conversation (void);
    register_tap_listener_ncp_conversation ();}
  { extern void register_tap_listener_rsvp_conversation (void);
    register_tap_listener_rsvp_conversation ();}
  { extern void register_tap_listener_sctp_conversation (void);
    register_tap_listener_sctp_conversation ();}
  { extern void register_tap_listener_tcpip_conversation (void);
    register_tap_listener_tcpip_conversation ();}
  { extern void register_tap_listener_tr_conversation (void);
    register_tap_listener_tr_conversation ();}
  { extern void register_tap_listener_udpip_conversation (void);
    register_tap_listener_udpip_conversation ();}
  { extern void register_tap_listener_usb_conversation (void);
    register_tap_listener_usb_conversation ();}
  { extern void register_tap_listener_wlan_conversation (void);
    register_tap_listener_wlan_conversation ();}
  { extern void register_tap_listener_gtkdcerpcstat (void);
    register_tap_listener_gtkdcerpcstat ();}
  { extern void register_tap_listener_gtkdiameterstat (void);
    register_tap_listener_gtkdiameterstat ();}
  { extern void register_tap_listener_expert_comp (void);
    register_tap_listener_expert_comp ();}
  { extern void register_tap_listener_expert (void);
    register_tap_listener_expert ();}
  { extern void register_tap_listener_gtkfcstat (void);
    register_tap_listener_gtkfcstat ();}
  { extern void register_tap_listener_flow_graph (void);
    register_tap_listener_flow_graph ();}
  { extern void register_tap_listener_gtkfunnel (void);
    register_tap_listener_gtkfunnel ();}
  { extern void register_tap_listener_gtkgsm_a_stat (void);
    register_tap_listener_gtkgsm_a_stat ();}
  { extern void register_tap_listener_gtkgsm_map_stat (void);
    register_tap_listener_gtkgsm_map_stat ();}
  { extern void register_tap_listener_gtkgsm_map_summary (void);
    register_tap_listener_gtkgsm_map_summary ();}
  { extern void register_tap_listener_gtkgtpstat (void);
    register_tap_listener_gtkgtpstat ();}
  { extern void register_tap_listener_gtk_h225counter (void);
    register_tap_listener_gtk_h225counter ();}
  { extern void register_tap_listener_gtk_h225rassrt (void);
    register_tap_listener_gtk_h225rassrt ();}
  { extern void register_tap_listener_eth_hostlist (void);
    register_tap_listener_eth_hostlist ();}
  { extern void register_tap_listener_fc_hostlist (void);
    register_tap_listener_fc_hostlist ();}
  { extern void register_tap_listener_fddi_hostlist (void);
    register_tap_listener_fddi_hostlist ();}
  { extern void register_tap_listener_ip_hostlist (void);
    register_tap_listener_ip_hostlist ();}
  { extern void register_tap_listener_ipx_hostlist (void);
    register_tap_listener_ipx_hostlist ();}
  { extern void register_tap_listener_jxta_hostlist (void);
    register_tap_listener_jxta_hostlist ();}
  { extern void register_tap_listener_ncp_hostlist (void);
    register_tap_listener_ncp_hostlist ();}
  { extern void register_tap_listener_rsvp_hostlist (void);
    register_tap_listener_rsvp_hostlist ();}
  { extern void register_tap_listener_sctp_hostlist (void);
    register_tap_listener_sctp_hostlist ();}
  { extern void register_tap_listener_tcpip_hostlist (void);
    register_tap_listener_tcpip_hostlist ();}
  { extern void register_tap_listener_tr_hostlist (void);
    register_tap_listener_tr_hostlist ();}
  { extern void register_tap_listener_udpip_hostlist (void);
    register_tap_listener_udpip_hostlist ();}
  { extern void register_tap_listener_usb_hostlist (void);
    register_tap_listener_usb_hostlist ();}
  { extern void register_tap_listener_wlan_hostlist (void);
    register_tap_listener_wlan_hostlist ();}
  { extern void register_tap_listener_iax2_analysis (void);
    register_tap_listener_iax2_analysis ();}
  { extern void register_tap_listener_gtk_iostat (void);
    register_tap_listener_gtk_iostat ();}
  { extern void register_tap_listener_gtkldapstat (void);
    register_tap_listener_gtkldapstat ();}
  { extern void register_tap_listener_mac_lte_stat (void);
    register_tap_listener_mac_lte_stat ();}
  { extern void register_tap_listener_mcast_stream_dlg (void);
    register_tap_listener_mcast_stream_dlg ();}
  { extern void register_tap_listener_gtkmegacostat (void);
    register_tap_listener_gtkmegacostat ();}
  { extern void register_tap_listener_gtkmgcpstat (void);
    register_tap_listener_gtkmgcpstat ();}
  { extern void register_tap_listener_gtkmtp3_stat (void);
    register_tap_listener_gtkmtp3_stat ();}
  { extern void register_tap_listener_gtkmtp3_summary (void);
    register_tap_listener_gtkmtp3_summary ();}
  { extern void register_tap_listener_gtkncpstat (void);
    register_tap_listener_gtkncpstat ();}
  { extern void register_tap_listener_gtkradiusstat (void);
    register_tap_listener_gtkradiusstat ();}
  { extern void register_tap_listener_gtkrpcprogs (void);
    register_tap_listener_gtkrpcprogs ();}
  { extern void register_tap_listener_gtkrpcstat (void);
    register_tap_listener_gtkrpcstat ();}
  { extern void register_tap_listener_rtp_analysis (void);
    register_tap_listener_rtp_analysis ();}
  { extern void register_tap_listener_rtp_stream_dlg (void);
    register_tap_listener_rtp_stream_dlg ();}
  { extern void register_tap_listener_gtkscsistat (void);
    register_tap_listener_gtkscsistat ();}
  { extern void register_tap_listener_sctp_analyse (void);
    register_tap_listener_sctp_analyse ();}
  { extern void register_tap_listener_sctpstat (void);
    register_tap_listener_sctpstat ();}
  { extern void register_tap_listener_sctp_stat_dlg (void);
    register_tap_listener_sctp_stat_dlg ();}
  { extern void register_tap_listener_gtksipstat (void);
    register_tap_listener_gtksipstat ();}
  { extern void register_tap_listener_gtksmbstat (void);
    register_tap_listener_gtksmbstat ();}
  { extern void register_tap_listener_gtksmb2stat (void);
    register_tap_listener_gtksmb2stat ();}
  { extern void register_tap_listener_stats_tree_stat (void);
    register_tap_listener_stats_tree_stat ();}
  { extern void register_tap_listener_t38_analysis (void);
    register_tap_listener_t38_analysis ();}
  { extern void register_tap_listener_tcp_graph (void);
    register_tap_listener_tcp_graph ();}
  { extern void register_tap_listener_voip_calls_dlg (void);
    register_tap_listener_voip_calls_dlg ();}
  { extern void register_tap_listener_wlanstat (void);
    register_tap_listener_wlanstat ();}
  { extern void register_tap_listener_gtkwspstat (void);
    register_tap_listener_gtkwspstat ();}
}
