; wsutil.def
; $Id: libwsutil.def 27077 2008-12-22 01:14:36Z guy $

; Add symbols that should be exported out of libwsutil.dll
;
; DATA symbols must be declared WS_VAR_IMPORT in the header file!
;
; NOTE: TABS DON'T WORK IN THIS FILE, USE SPACES INSTEAD!!!
;
EXPORTS

; file_util.c
ws_stdio_fopen
ws_stdio_freopen
ws_stdio_mkdir
ws_stdio_open
ws_stdio_remove
ws_stdio_rename
ws_stdio_stat
ws_stdio_unlink

; mpeg-audio.c
mpa_bitrate
mpa_frequency
mpa_layer
mpa_padding
mpa_samples
mpa_version

; privileges.c
get_credential_info
get_cur_groupname
get_cur_username
npf_sys_is_running
relinquish_special_privs_perm
running_with_special_privs
started_with_special_privs

; str_util.c
ascii_strdown_inplace
ascii_strup_inplace

; type_util.c
type_util_gdouble_to_guint64
type_util_guint64_to_gdouble

; unicode-utils.c
utf_16to8
utf_8to16

